@@ -2,7 +2,7 @@
 Utility module to manipulate queries
 """
 
-# Copyright (C) 2020-2021 The Psycopg Team
+# Copyright (C) 2020 The Psycopg Team
 
 import re
 from typing import Any, Dict, List, Mapping, Match, NamedTuple, Optional
@@ -14,7 +14,7 @@ from . import errors as e
 from .sql import Composable
 from .abc import Buffer, Query, Params
 from ._enums import PyFormat
-from ._encodings import pgconn_encoding
+from ._encodings import conn_encoding
 
 if TYPE_CHECKING:
     from .abc import Transformer
@@ -47,15 +47,11 @@ class PostgresQuery:
         self._want_formats: Optional[List[PyFormat]] = None
         self.formats: Optional[Sequence[pq.Format]] = None
 
+        self._encoding = conn_encoding(transformer.connection)
         self._parts: List[QueryPart]
         self.query = b""
-        self._encoding = "utf-8"
         self._order: Optional[List[str]] = None
 
-        conn = transformer.connection
-        if conn:
-            self._encoding = pgconn_encoding(conn.pgconn)
-
     def convert(self, query: Query, vars: Optional[Params]) -> None:
         """
         Set up the query and parameters to convert.
@@ -90,9 +86,7 @@ class PostgresQuery:
         This method updates `params` and `types`.
         """
         if vars is not None:
-            params = _validate_and_reorder_params(
-                self._parts, vars, self._order
-            )
+            params = _validate_and_reorder_params(self._parts, vars, self._order)
             assert self._want_formats is not None
             self.params = self._tx.dump_sequence(params, self._want_formats)
             self.types = self._tx.types or ()
@@ -103,6 +97,53 @@ class PostgresQuery:
             self.formats = None
 
 
+class PostgresClientQuery(PostgresQuery):
+    """
+    PostgresQuery subclass merging query and arguments client-side.
+    """
+
+    __slots__ = ("template",)
+
+    def convert(self, query: Query, vars: Optional[Params]) -> None:
+        """
+        Set up the query and parameters to convert.
+
+        The results of this function can be obtained accessing the object
+        attributes (`query`, `params`, `types`, `formats`).
+        """
+        if isinstance(query, str):
+            bquery = query.encode(self._encoding)
+        elif isinstance(query, Composable):
+            bquery = query.as_bytes(self._tx)
+        else:
+            bquery = query
+
+        if vars is not None:
+            (self.template, self._order, self._parts) = _query2pg_client(
+                bquery, self._encoding
+            )
+        else:
+            self.query = bquery
+            self._order = None
+
+        self.dump(vars)
+
+    def dump(self, vars: Optional[Params]) -> None:
+        """
+        Process a new set of variables on the query processed by `convert()`.
+
+        This method updates `params` and `types`.
+        """
+        if vars is not None:
+            params = _validate_and_reorder_params(self._parts, vars, self._order)
+            self.params = tuple(
+                self._tx.as_literal(p) if p is not None else b"NULL" for p in params
+            )
+            self.query = self.template % self.params
+        else:
+            self.params = None
+
+
 @lru_cache()
 def _query2pg(
     query: bytes, encoding: str
@@ -112,7 +153,7 @@ def _query2pg(
 
     - Convert Python placeholders (``%s``, ``%(name)s``) into Postgres
       format (``$1``, ``$2``)
-    - placeholders can be %s or %b (text or binary)
+    - placeholders can be %s, %t, or %b (auto, text or binary)
     - return ``query`` (bytes), ``formats`` (list of formats) ``order``
       (sequence of names used in the query, in the position they appear)
       ``parts`` (splits of queries and placeholders).
@@ -144,8 +185,7 @@ def _query2pg(
             else:
                 if seen[part.item][1] != part.format:
                     raise e.ProgrammingError(
-                        f"placeholder '{part.item}' cannot have"
-                        f" different formats"
+                        f"placeholder '{part.item}' cannot have different formats"
                     )
                 chunks.append(seen[part.item][0])
 
@@ -155,6 +195,43 @@ def _query2pg(
     return b"".join(chunks), formats, order, parts
 
 
+@lru_cache()
+def _query2pg_client(
+    query: bytes, encoding: str
+) -> Tuple[bytes, Optional[List[str]], List[QueryPart]]:
+    """
+    Convert Python query and params into a template to perform client-side binding
+    """
+    parts = _split_query(query, encoding)
+    order: Optional[List[str]] = None
+    chunks: List[bytes] = []
+
+    if isinstance(parts[0].item, int):
+        for part in parts[:-1]:
+            assert isinstance(part.item, int)
+            chunks.append(part.pre)
+            chunks.append(b"%s")
+
+    elif isinstance(parts[0].item, str):
+        seen: Dict[str, Tuple[bytes, PyFormat]] = {}
+        order = []
+        for part in parts[:-1]:
+            assert isinstance(part.item, str)
+            chunks.append(part.pre)
+            if part.item not in seen:
+                ph = b"%s"
+                seen[part.item] = (ph, part.format)
+                order.append(part.item)
+                chunks.append(ph)
+            else:
+                chunks.append(seen[part.item][0])
+
+    # last part
+    chunks.append(parts[-1].pre)
+
+    return b"".join(chunks), order, parts
+
+
 def _validate_and_reorder_params(
     parts: List[QueryPart], vars: Params, order: Optional[List[str]]
 ) -> Sequence[Any]:
@@ -173,7 +250,7 @@ def _validate_and_reorder_params(
         sequence = False
     else:
         raise TypeError(
-            f"query parameters should be a sequence or a mapping,"
+            "query parameters should be a sequence or a mapping,"
             f" got {type(vars).__name__}"
         )
 
@@ -184,9 +261,7 @@ def _validate_and_reorder_params(
                 f" {len(vars)} parameters were passed"
             )
         if vars and not isinstance(parts[0].item, int):
-            raise TypeError(
-                "named placeholders require a mapping of parameters"
-            )
+            raise TypeError("named placeholders require a mapping of parameters")
         return vars  # type: ignore[return-value]
 
     else:
@@ -195,12 +270,10 @@ def _validate_and_reorder_params(
                 "positional placeholders (%s) require a sequence of parameters"
             )
         try:
-            return [
-                vars[item] for item in order or ()  # type: ignore[call-overload]
-            ]
+            return [vars[item] for item in order or ()]  # type: ignore[call-overload]
         except KeyError:
             raise e.ProgrammingError(
-                f"query parameter missing:"
+                "query parameter missing:"
                 f" {', '.join(sorted(i for i in order or () if i not in vars))}"
             )
 
@@ -257,7 +330,7 @@ def _split_query(query: bytes, encoding: str = "ascii") -> List[QueryPart]:
 
         if ph == b"%(":
             raise e.ProgrammingError(
-                f"incomplete placeholder:"
+                "incomplete placeholder:"
                 f" '{query[m.span(0)[0]:].split()[0].decode(encoding)}'"
             )
         elif ph == b"% ":
@@ -268,7 +341,7 @@ def _split_query(query: bytes, encoding: str = "ascii") -> List[QueryPart]:
             )
         elif ph[-1:] not in b"sbt":
             raise e.ProgrammingError(
-                f"only '%s', '%b', '%t' are allowed as placeholders, got"
+                "only '%s', '%b', '%t' are allowed as placeholders, got"
                 f" '{m.group(0).decode(encoding)}'"
             )
 
